/*
 * encoder.c for miniD pre OLED
 *
 * Rotary encoder functions including encoder rotation and button push.
 *
 * Initial version: 20/12/2017
 * Author: Tomi Nihtila, nihtila.com
 */ 


#include "encoder.h"
#include "main.h"
#include "globals.h"
#include "audio.h"
#include "eeprom.h"
#include <avr/io.h>
#include <stdbool.h>
#include <stdlib.h>


// encoder object with relevant data
encoderSt encoder = {
	{0,1,-1,0,-1,0,0,1,1,0,0,-1,0,-1,1,0},	// State LUT: 0 means stationary, -1 counter-clockwise and 1 clockwise rotation
	0,										// ab
	0,										// counter
	HIGH,									// butState
	0,										// countBut
	50/(ENC_BUT_CHECK_INTERVAL*TICK),		// intervalShortBut
	1500/(ENC_BUT_CHECK_INTERVAL*TICK)		// intervalLongBut
};



// ********************************************************
// ROTATION ISR
// This tracks encoder button pushing and is called inside ISR, therefore it should not dive deep
// into function calls but rather call events.
//
void encoderButtonISR(void) {
	
	// If button is now low and was high - it was just pushed
	if(((PIND & 0x08) == 0) && (encoder.butState == HIGH)) {

		encoder.butState = LOW;

		// Set up event to count the length of button push
		resetEvent(BUTTON_PUSH_EV, true, false);
	}

	// If button is now high but it was low last time - button release
	if(((PIND & 0x08) != 0) && (encoder.butState == LOW)) {
	
		// stop counting button push duration
		resetEvent(BUTTON_PUSH_EV, false, false);

		// Was it pushed long enough to recognised as short button push? (long push is handled elsewhere)
		if((encoder.countBut > encoder.intervalShortBut) && (encoder.countBut < encoder.intervalLongBut)) {

			resetEvent(SH_BUTTON_EV, true, true);
		}

		// Reset button push counter and button state
		encoder.countBut = 0;
		encoder.butState = HIGH;
	}	
}


// ********************************************************
// ROTATION ISR
// This tracks encoder rotation and is called inside ISR, therefore it should not dive deep
// into function calls.
//
void encoderRotateISR(unsigned char abOld) {

	unsigned char abSum;	
	
	// Sum holds previous value at bits 3-2 and current value at bits 1-0
	abSum = abOld << 2;
	abSum += encoder.ab;
		
	// Some encoders seem to work opposite way
	#ifndef ENC_REVERSE
		// Encoder look-up table can be indexed directly using that value
		encoder.counter += encoder.luTable[abSum];
	#else
		encoder.counter -= encoder.luTable[abSum];
	#endif
	
	// If enough change has been detected, call encoder change event
	if(encoder.counter > 3 || encoder.counter < -3) {
		resetEvent(ENCODER_CHANGE_EV, true, true);
	}	
}


// ********************************************************
// ROTARY ENCODER CHANGE
// Changes data values based on encoder rotation. This is called as an event function
// when rotation has been confirmed.
//
void encoderChange(void) {

	signed char change;


	// If event was resetted at the end as usually, there is a possibility some rotation
	// during this function would be ignored
	resetEvent(ENCODER_CHANGE_EV, false, false);

	// VOLUME AND INPUT DISPLAY - change volume
	if(sys.state == VOLUME_ST) {

		change = encoder.counter/4;

		preamp.masterVol += change;
		
		// Check volume value validity and update if necessary
		audioUpdateInput();
		audioUpdateVolume();
	}
	
	// MENU - change active menu item
	else if(sys.state == MENU_ST) {
		
		if(encoder.counter > 0) {
			if(sys.menuIndex < sys.noMenuItems-1)
			sys.menuIndex++;
		}
		else if(encoder.counter < 0) {
			if(sys.menuIndex > 0)
			sys.menuIndex--;
		}
	}

	// MENU ITEM SELECTED - change setting value
	else if(sys.state == MENUITEM_ST) {
		
		if(encoder.counter > 0) {
			if(sys.menuItem[sys.menuIndex].data < sys.menuItem[sys.menuIndex].dataMax) {
				sys.menuItem[sys.menuIndex].data++;
			}
		}
		else if(encoder.counter < 0) {
			if(sys.menuItem[sys.menuIndex].data > sys.menuItem[sys.menuIndex].dataMin) {
				sys.menuItem[sys.menuIndex].data--;
			}
		}
		
		// Call function to handle the change of data
		if(sys.menuItem[sys.menuIndex].dataFunc != NULL) {
			sys.menuItem[sys.menuIndex].dataFunc();
		}

		sys.dispBlink = false;
	}

	encoder.counter = 0;
	resetEvent(DISP_REFRESH_EV, true, true);
}


// ********************************************************
// ENCODER TIMEOUT
// This relates to smoother rotary encoder use. Sometimes rotation kind of loses track slightly
// and change is not registered at detents but somewhere between. This resets rotation counter
// when shaft has been at detent for a short period of time.
//
void encoderTimeout(void) {
	
	if(((PIND & 0x30) >> 4) == 0x03) {
		encoder.counter = 0;
	}
	
	resetEvent(ENCODER_TIMEOUT_EV, false, false);
}



// ********************************************************
// BUTTON PUSH DURATION COUNTER
// Keeps track of button push duration, used as event function and called frequently.
//
void encoderButtonPushCount(void) {
	
	// If button is pushed, count the time it is being pushed
	if(encoder.butState == LOW) {

		encoder.countBut++;

		// If it has been pushed long enough to register long button push
		if(encoder.countBut == encoder.intervalLongBut) {

			encoderLongButtonPush();

			// This keeps button push counter at constant value before user releases it
			// to avoid registering multiple pushes
			encoder.countBut = encoder.intervalLongBut;
		}
	}

	// Keep this event occurring until button is released (controlled in encoderButtonISR)
	resetEvent(BUTTON_PUSH_EV, true, false);
}


// ********************************************************
// SHORT BUTTON PUSH
// Called when short button push has been confirmed.
//
void encoderShortButtonPush(void) {

	// In normal volume/input display button push changes input,
	// also works in error state
	if((sys.state == VOLUME_ST) || (sys.state == ERROR_ST)) {
		preamp.input += 1;
    
		// Check validity of new value and change input
		audioUpdateInput();
	}

	// If at menu, button push selects menu item
	else if(sys.state == MENU_ST) {
		sys.state = MENUITEM_ST;
		sys.dispBlink = false;
	}

	// If menu item is selected, button push de-selects it
	else if(sys.state == MENUITEM_ST) {
		sys.state = MENU_ST;
		sys.dispBlink = false;
	}
  
	resetEvent(SH_BUTTON_EV, false, false);
	resetEvent(DISP_REFRESH_EV, true, true);
}


// ********************************************************
// LONG BUTTON PUSH
// Called when long button push has been confirmed.
//
void encoderLongButtonPush(void) {

	// Long button push switches between normal display and menu
	if(sys.state == VOLUME_ST) {
		sys.state = MENU_ST;
		sys.menuIndex = 0;
	}
	else {
		sys.state = VOLUME_ST;
		// Save settings when exiting menu
		saveToEEPROM();
	}
	
	resetEvent(DISP_REFRESH_EV, true, true);
}
