/*
 * main.c for miniD pre OLED
 *
 * Includes global system and preamp structure initialisations, menu string definitions. Code
 * includes all data and peripheral initialisations, interrupt service routines, event handling,
 * startup function and display update main function. Main code loop contains event handles
 * (simple scheduler).
 *
 * Initial version: 27/12/2017
 * Author: Tomi Nihtila, nihtila.com
 */



#include "main.h"
#include "globals.h"
#include "audio.h"
#include "eeprom.h"
#include "ir.h"
#include "encoder.h"
#include "oled.h"
#include <avr/io.h>
#include <stdbool.h>
#include <avr/interrupt.h>
#include <stdlib.h>


// FUNCTION DECLARATIONS
// Define functions which are needed before their definition and in struct initialisations
static void dispWrite(void);
static void startup(void);


// MENU ITEM TEXTS
// With current font and settings 11 characters is the maximum
const char MENU_T_DSP_BYP[] PROGMEM =	"DSP bypass";
const char MENU_T_OUT2_MUTE[] PROGMEM =	"Out2 mute";
const char MENU_T_OUT2_OFFS[] PROGMEM =	"Out2 offset";
const char MENU_T_VOL_CONT[] PROGMEM =	"Volume cont";
const char MENU_T_USE_OP2[] PROGMEM =	"Use Op2 in";
const char MENU_T_USE_CX2[] PROGMEM =	"Use Cx2 in";
const char MENU_T_USE_ANA[] PROGMEM =	"Use Ana in";
const char MENU_T_START_VOL[] PROGMEM =	"Startup vol";
const char MENU_T_VOL_LIMIT[] PROGMEM =	"Vol limit";
const char MENU_T_OP1_OFFS[] PROGMEM =	"Op1 offset";
const char MENU_T_OP2_OFFS[] PROGMEM =	"Op2 offset";
const char MENU_T_CX1_OFFS[] PROGMEM =	"Cx1 offset";
const char MENU_T_CX2_OFFS[] PROGMEM =	"Cx2 offset";
const char MENU_T_ANA_OFFS[] PROGMEM =	"Ana offset";
const char MENU_T_OUT1_BAL[] PROGMEM =	"Out1 bal";
const char MENU_T_OUT2_BAL[] PROGMEM =	"Out2 bal";
const char MENU_T_START_DEL[] PROGMEM =	"Startup del";
const char MENU_T_SRC_F_GD[] PROGMEM =	"SRC filt GD";
const char MENU_T_DEEMP_F[] PROGMEM =	"De-emp filt";
const char MENU_T_LEARN_IR[] PROGMEM =	"Learn IR";

// MENU ITEM UNITS
const char MENU_U_DB[] PROGMEM =		"dB";
const char MENU_U_SECOND[] PROGMEM =	"s";
const char MENU_U_SAMPLE[] PROGMEM =	"samples";


// MENU ITEM OPTIONS
// Replaces numerical data in some menu items

// Yes/no
const char MENU_ST_YES[] PROGMEM =		"Yes";
const char MENU_ST_NO[] PROGMEM =		"No";
// De-emphasis filter
const char MENU_ST_DISABLED[] PROGMEM =	"Disabled";
const char MENU_ST_FOR48K[] PROGMEM =	" For 48k";
const char MENU_ST_FOR44K[] PROGMEM =	"For 44.1k";
const char MENU_ST_FOR32K[] PROGMEM =	" For 32k";
const char MENU_ST_AUTO[] PROGMEM =		" Auto";
// Analog/digital
const char MENU_ST_ANALOG[] PROGMEM =	"Analog";
const char MENU_ST_DIGITAL[] PROGMEM =	"Digital";
// SRC filter group delay
const char MENU_ST_64[] PROGMEM =		"64";
const char MENU_ST_32[] PROGMEM =		"32";
const char MENU_ST_16[] PROGMEM =		"16";
const char MENU_ST_8[] PROGMEM =		"8";
// IR commands
const char MENU_ST_UP[] PROGMEM =		"  Up";
const char MENU_ST_DOWN[] PROGMEM =		" Down";
const char MENU_ST_SHBUT[] PROGMEM =	"Sh but";
const char MENU_ST_LNBUT[] PROGMEM =	"Ln but";
// Balance
const char MENU_ST_BAL_0[] PROGMEM =	"L      -5dB";
const char MENU_ST_BAL_1[] PROGMEM =	" L     -4dB";
const char MENU_ST_BAL_2[] PROGMEM =	"  L    -3dB";
const char MENU_ST_BAL_3[] PROGMEM =	"   L   -2dB";
const char MENU_ST_BAL_4[] PROGMEM =	"    L  -1dB";
const char MENU_ST_BAL_5[] PROGMEM =	"   Center  ";
const char MENU_ST_BAL_6[] PROGMEM =	"-1dB  R    ";
const char MENU_ST_BAL_7[] PROGMEM =	"-2dB   R   ";
const char MENU_ST_BAL_8[] PROGMEM =	"-3dB    R  ";
const char MENU_ST_BAL_9[] PROGMEM =	"-4dB     R ";
const char MENU_ST_BAL_10[] PROGMEM =	"-5dB      R";




// Options for each menu item, NULL = numerical value without text mapping
// Note that underlying data is still numerical and order matters! For example, if 0="no" and 1="yes"
// give the options in this order. Moreover, data must be limited to 0-1 in struct initialization.
// These tables are simply used for mapping using indexing, no validity checks are done.
// const char* const means both pointer and data are constant
const char* const MENU_O_NO_YES[] PROGMEM =		{MENU_ST_NO, MENU_ST_YES};
const char* const MENU_O_VOL_CONT[] PROGMEM =	{MENU_ST_ANALOG, MENU_ST_DIGITAL};
const char* const MENU_O_SRC_F_GD[] PROGMEM =	{MENU_ST_64, MENU_ST_32, MENU_ST_16, MENU_ST_8};
const char* const MENU_O_DEEMP_F[] PROGMEM =	{MENU_ST_DISABLED, MENU_ST_FOR48K, MENU_ST_FOR44K, MENU_ST_FOR32K, MENU_ST_AUTO};
const char* const MENU_O_LEARN_IR[] PROGMEM =	{MENU_ST_UP, MENU_ST_DOWN, MENU_ST_SHBUT, MENU_ST_LNBUT};
const char* const MENU_O_BALANCE[] PROGMEM =	{MENU_ST_BAL_0, MENU_ST_BAL_1, MENU_ST_BAL_2, MENU_ST_BAL_3,
												 MENU_ST_BAL_4, MENU_ST_BAL_5, MENU_ST_BAL_6, MENU_ST_BAL_7,
												 MENU_ST_BAL_8, MENU_ST_BAL_9, MENU_ST_BAL_10};
// Data X-location
const unsigned char MENU_XLOC[] PROGMEM =		{3, 15, 25};



sysSt sys = {
	STARTUP_ST,                           // Initial state is startup state
	0,                                    // Event timer starts at 0
	false,                                // Display blinking off
	0,                                    // Initial menu index
	NO_MENUITEMS, {	// Initialise menu items. Note that default values are replaced with EEPROM saved values at startup. In some cases
					// also min/max values may be changed at runtime.
					// NOTE THAT THIS ORDER MUST MATCH THE ORDER OF MENUITEM ENUM IN GLOBALS.H!
		//menu text,		init data,	data min,	data max,	unit string,	option strings,	data x-loc,		func to call when data is changed
		{MENU_T_DSP_BYP,	0,			0,			1,			NULL,			MENU_O_NO_YES,	&MENU_XLOC[2],	&audioDSPBypassControl},
		{MENU_T_OUT2_MUTE,	0,			0,			1,			NULL,			MENU_O_NO_YES,	&MENU_XLOC[2],	&audioMuteControl},
		{MENU_T_OUT2_OFFS,	0,			MIN_OFFSET,	MAX_OFFSET, MENU_U_DB,		NULL,			&MENU_XLOC[2],	&audioUpdateVolume},
		{MENU_T_VOL_CONT,	0,			0,			1,			NULL,			MENU_O_VOL_CONT,&MENU_XLOC[2],	&audioChangeVolumeControl},
		{MENU_T_USE_OP2,	1,			0,			1,			NULL,			MENU_O_NO_YES,	&MENU_XLOC[2],	&audioUpdateInput},
		{MENU_T_USE_CX2,	1,			0,			1,			NULL,			MENU_O_NO_YES,	&MENU_XLOC[2],	&audioUpdateInput},
		{MENU_T_USE_ANA,	1,			0,			1,			NULL,			MENU_O_NO_YES,	&MENU_XLOC[2],	&audioUpdateInput},
		{MENU_T_START_VOL,	DEF_GAIN,	MIN_GAIN,	MAX_GAIN,	MENU_U_DB,		NULL,			&MENU_XLOC[2],	NULL},
		{MENU_T_VOL_LIMIT,	MAX_GAIN,	MIN_GAIN,	MAX_GAIN,	MENU_U_DB,		NULL,			&MENU_XLOC[2],	&audioUpdateVolume},
		{MENU_T_OP1_OFFS,	0,			MIN_OFFSET, MAX_OFFSET, MENU_U_DB,		NULL,			&MENU_XLOC[2],	&audioUpdateVolume},
		{MENU_T_OP2_OFFS,	0,			MIN_OFFSET, MAX_OFFSET, MENU_U_DB,		NULL,			&MENU_XLOC[2],	&audioUpdateVolume},
		{MENU_T_CX1_OFFS,	0,			MIN_OFFSET, MAX_OFFSET, MENU_U_DB,		NULL,			&MENU_XLOC[2],	&audioUpdateVolume},
		{MENU_T_CX2_OFFS,	0,			MIN_OFFSET, MAX_OFFSET, MENU_U_DB,		NULL,			&MENU_XLOC[2],	&audioUpdateVolume},
		{MENU_T_ANA_OFFS,	0,			MIN_OFFSET, MAX_OFFSET, MENU_U_DB,		NULL,			&MENU_XLOC[2],	&audioUpdateVolume},
		{MENU_T_OUT1_BAL,	5,			0,			10,			NULL,			MENU_O_BALANCE,	&MENU_XLOC[0],	&audioUpdateVolume},
		{MENU_T_OUT2_BAL,	5,			0,			10,			NULL,			MENU_O_BALANCE,	&MENU_XLOC[0],	&audioUpdateVolume},
		{MENU_T_START_DEL,	2,			0,			5,			MENU_U_SECOND,	NULL,			&MENU_XLOC[2],	NULL},
		{MENU_T_SRC_F_GD,	0,			0,			3,			MENU_U_SAMPLE,	MENU_O_SRC_F_GD,&MENU_XLOC[0],	&audioSRCChangeSettings},
		{MENU_T_DEEMP_F,	0,			0,			4,			NULL,			MENU_O_DEEMP_F,	&MENU_XLOC[1],	&audioSRCChangeSettings},
		{MENU_T_LEARN_IR,	0,			0,			NO_IR_CMD-1,NULL,			MENU_O_LEARN_IR,&MENU_XLOC[1],	NULL}},
		
		// Initialise events, times are given in ms/TICK, ms is the time in milliseconds. Note that startup delay is also a menu
		// variable and is replaced with the EEPROM value at startup.
		// (0/TICK in interval basically means current event is called to execute immediately)
		// time of 1st exec,	execute,	interval,				event function
		{{2000/TICK,			true,		0/TICK,					&startup},					// startup delay
		{0/TICK,				false,		0/TICK,					&encoderChange},			// encoder position change
		{0/TICK,				false,		22/TICK,				&encoderTimeout},			// encoder timeout
		{0/TICK,				false,		0/TICK,					&encoderShortButtonPush},	// encoder short button push
		{0/TICK,				false,		ENC_BUT_CHECK_INTERVAL,	&encoderButtonPushCount},	// encoder button push counter
		{0/TICK,				false,		0/TICK,					&irAction},					// IR action
		{0/TICK,				false,		160/TICK,				&irTimeout},				// IR timeout
		{0/TICK,				false,		230/TICK,				&audioSRCInt},				// SRC interrupt	
		{0/TICK,				false,		500/TICK,				&dispWrite}},				// display refresh
		
		// IR data initialisation, codes will be read from EEPROM if stored
		{false,				// repeatCode
		false,				// timeout
		NO_IR_CMD,			// no of commants
		0,					// code
		{0, 0, 0, 0}},		// saved codes
			
		// error code
		0x00
	};


// Preamp initialisation
// Some value are changed at startup with EEPROM-saved values.
preampSt preamp = {
	DEF_GAIN,                           // Initial gain, changed at startup
	MIN_GAIN-1,							// mingain-1 = mute
	MAX_GAIN,							// Will be replaced by menu-set maximum volume
	0,                                  // Input, changed at startup
	true,                               // Mute on at startup
	NO_INPUTS,                          // Number of inputs
	{"O1", "O2", "C1", "C2", "An"},		// input names
	true,								// signal lock
	0
};




int main(void)
{
	// INITIALISATIONS
	unsigned char i;

	// Originally used for LCD, not used here (except for debugging)
	PORTA = 0xFF;
	DDRA = 0xFF;
	
	// For SPI, CSs and LCD backlight
	// clk, miso, mosi, cs_pga, backlight, cs_dac, cs_src, src_int
	DDRB =	0b10111110;
	// All high except backlight pin, src_int has pullup on board
	PORTB = 0b10111110;
	
	// PORTC is for JTAG and handled by the MCU except I2C
	DDRC = 0x03;
	PORTC = 0x03;
	
	// outp_mute (on), ir, rotenc1, rotenc2, rotbut, dsp_sel2, dsp_sel1, nc
	DDRD =	0b10000111;
	PORTD = 0b00111000;
	
	// Comparator off
	PCICR = 0x80;


	// Display init
	oledInit();
	
	  
	// LOAD SAVED SETTINGS
	// All settings (input, and menu settings) are saved in EEPROM as defaults when exiting menu.
	readEEPROM();
	
	// Set startup volume
	preamp.mute = true;	
	preamp.masterVol = sys.menuItem[STARTUP_VOL_MI].data;	
	
	// Initialise SRC and DAC (as well as SPI)
	audioInit();	

	// Keep mute on
	audioMuteControl();

	// Set display update call when initialisations are ready
	resetEvent(DISP_REFRESH_EV, true, true);

	// Set startup delay (for calling startup()). At startup() function mute is released, 
	// volume updated, and rotary encoder operation initialised.
	sys.events[STARTUP_EV].ticks = (unsigned int)sys.menuItem[STARTUP_DEL_MI].data*1000/TICK;

	  
	// Set timer2 to trig output compare interrupt every tick (default 2 ms, not accurate).
	// Note that IR uses the same counter for delays and the longest IR delay is 9ms, therefore
	// T2 overflow must be longer than that (i.e. prescaler must be 1024 which leads to rather
	// poor accuracy for tick).
	OCR2A = 16;			// 16x 128 us = 2048 us
	TIMSK2 = 0x02;		// enable output compare A interrupt
	TCNT2 = 0;
	TCCR2B = 0x07;		// timer on, prescaler to 1024 (one timer tick is 128 us)
	  
	// enable interrupts
	sei();
	
	// Rotary encoder and pin change interrupts are initialised at startup() function
	
	
	// Main loop takes care of calling event functions (simple scheduler), checking rotary encoder
	// changes (handled by ISRs), and checking if IR code has been received (handled by ISR).
    while (1) 
    {

		// This is the basic scheduler loop which constantly checks if an event needs to be executed.
		// This is signaled by 'execute' flag and 'ticks' which is a local timebase. Event functions
		// can be called by using resetEvent().
		for(i = 0; i < NO_EVENTS; i++) {
			
			// 'Smaller than' is included so that event execution will not be missed if one round here
			// is missed. However, this causes issue close to counter overflow. For example, if 
			// sys.ticks is 65500 and event.interval is 500, event.ticks gets value 465. This is smaller
			// than 65500 so this events is going to be executed many times before sys.ticks overflows
			// as well. The last check of AND clause ensures MSB of both ticks is equal. 
			if((sys.events[i].ticks <= sys.ticks) && (sys.events[i].execute == true) && 
			!((sys.events[i].ticks ^ sys.ticks) & 0x8000)) {
		  
				sys.events[i].func();
									
			}
		}
    }
}



// ********************************************************
// SCHEDULER - TIMER2 OUTPUT COMPARE INTERRUPT
// Occurs every 2 ms and keeps track of all timings. Actual event call is handled in main loop.
//
ISR(TIMER2_COMPA_vect) {

	// Free running unsigned int counter, if one tick is 10ms it takes around 10mins to overflow
	sys.ticks++;

	// Reset counter
	TCNT2 = 0;
}



// ********************************************************
// ROTARY ENCODER
// ********************************************************

// PIN CHANGE INTERRUPT
// Registers rotary encoder rotation and button push
//
ISR(PCINT3_vect) {

	static unsigned char abOld;
	static unsigned char portState = 0xFF;
	
	// PCINT may be caused by rotenc or IR - check if IR pin has changed value since last time
	// Xor gives 1 if value has changed
	if(((PIND ^ portState) & 0x40) != 0) {
		IRISR();
	}
	
	portState = PIND;

	abOld = encoder.ab;
	encoder.ab = (PIND & 0x30) >> 4;

	// BUTTON PUSH
	// If encoder state is unchanged, interrupt was caused by the button
	if(abOld == encoder.ab) {

		encoderButtonISR();
	}
	// ENCODER ROTATION
	else {
		encoderRotateISR(abOld);
	}

	resetEvent(ENCODER_TIMEOUT_EV, true, false);
}



// ********************************************************
// SRC PIN CHANGE INTERRUPT
// Sets SRC interrupt event.
//
ISR(PCINT1_vect) {

	// Call SRC interrupt check (if not in analog input)
	if(preamp.input < 4) {
		resetEvent(SRC_INT_EV, true, true);
	}
}
	

// ********************************************************
// DISPLAY UPDATE
// Event function that calls OLED update.
//
static void dispWrite(void) {

	sys.events[DISP_REFRESH_EV].execute = false;
		
	oledUpdate();

	// blink control - blinking in volume state if signal unlock, or when changing menu item
	if((sys.state == MENUITEM_ST) || ((preamp.signalLock == false) && (sys.state == VOLUME_ST))) {
		if(sys.dispBlink == false) {
			sys.dispBlink = true;
		}
		else {
			sys.dispBlink = false;
		}
		// call display update after set interval to keep blinking
		resetEvent(DISP_REFRESH_EV, true, false);
	}

}


// ********************************************************
// EVENT RESET
// After event is executed it is resetted, either not to execute anymore or to execute again after a
// time interval. This is not an actual event function but sets up an execution of one.
//
void resetEvent(eventNames event, bool execute, bool executeNow) {

	// If execute is set to false, .ticks value does not actually matter as it will not be executed anyway
	if(executeNow == false)
		sys.events[event].ticks = sys.ticks + sys.events[event].interval;
	else
		sys.events[event].ticks = sys.ticks;
	
	sys.events[event].execute = execute;
}


// ********************************************************
// STARTUP
// After set startup delay, mute is released and volume/input shown on display.
//
static void startup(void) {
	
	// Change system state
	sys.state = VOLUME_ST;

	// Enable rotary encoder operation
	// Initial state
	encoder.ab = (PIND & 0x30) >> 4;

	// Enable pin change interrupts on rotary encoder and IR
	//PCMSK3 = 0x38;	// rotenc only
	PCMSK3 = 0x78;	// rotenc and ir
	// Enable interrupt on SRC INT pin
	PCMSK1 = 0x01;
	PCICR = 0x0A;

	// Disable this startup event
	resetEvent(STARTUP_EV, false, false);

	// Call for SRC status check
	resetEvent(SRC_INT_EV, true, true);
	
	// Select input
	audioUpdateInput();
	
	// Update volume (releases mute)
	audioUpdateVolume();

	// Call for display update
	resetEvent(DISP_REFRESH_EV, true, false);
}




